module Main

import System
import System.Path
import System.File
import System.Directory
import System.Console.GetOpt
import Data.String
import Data.Vect

import Control.Eff
import Derive.Prelude

import Structures.Dependent.FreshList

import Runner
import Util
import Util.Eff

import Years.Y2015;

%language ElabReflection
%default total

data Flag = UseExample | Verbose
%runElab derive "Flag" [Eq, Show]

advent : Advent
advent = MkAdvent [
    y2015
  ]

options : List (OptDescr Flag)
options =
  [
    MkOpt ['u'] ["use-example"] (NoArg UseExample) "Use example input instead of main input"
  , MkOpt ['v'] ["verbose"] (NoArg Verbose) "Enable logging to stderr"
  ]

header : String
header = "Usage: advent YEAR DAY [OPTION...]"

data Error : Type where
  OptionsError : (errs : List String) -> Error
  ArgumentsError : (count : Nat) -> Error
  InvalidYear : (year : String) -> Error
  InvalidDay : (year : Nat) -> (day : String) -> Error
  NoSuchDay : (year : Nat) -> (day : Nat) -> Error
  NoCurrentDir : Error
  InputRead : (path : String) -> (err : FileError) -> Error
  SolveError : Show e => (year, day, part : Nat) -> (err : e) -> Error
%name Error err

Show Error where
  show (OptionsError errs) =
    let errs = unlines . map ("  " ++) . lines . joinBy "\n" $ errs
    in "Error parsing options:\n" ++ errs ++ "\n" ++ usageInfo header options
  show (ArgumentsError count) =
    "Error parsing arguments: Expected 2 positional arguments, got \{show count}. Provide only year and date.\n" ++
    usageInfo header options
  show (InvalidYear year) =
    "Error parsing arguments: Unable to parse year argument: \{year}"
  show (InvalidDay year day) =
    "Error parsing arguments: Failed to parse day \{day} of year \{show year}"
  show (NoSuchDay year day) =
    "Error locating day: No such day \{show day} of year \{show year}"
  show NoCurrentDir =
    "Unknown error getting current working directory"
  show (InputRead path err) =
    let err = unlines . map ("  " ++) . lines . show $ err
    in "Error reading input: Encountered error reading input at \{path}\n" ++ err
  show (SolveError year day part err) =
    let err = unlines . map ("  " ++) . lines . show $ err
    in "Error solving day \{show day} part \{show part} of year \{show year}: \n" ++ err

||| Convert the non-option arguments into a Year/Day pair
argumentsToPair : Has (Except Error) fs =>
  List String -> Eff fs (Nat, Nat)
argumentsToPair [] = throw $ ArgumentsError 0
argumentsToPair [x] = throw $ ArgumentsError 1
argumentsToPair [year, day] = do
  year <- note (InvalidYear year) $ parsePositive year
  day <- note (InvalidDay year day) $ parsePositive day
  pure (year, day)
argumentsToPair xs = throw $ ArgumentsError (length xs)

covering
||| Actual main, as an effectful computation
start : Eff [IO, Except Error] ()
start = do
  -- Read and parse arguments/options
  args <- getArgs
  let opts = getOpt Permute options (drop 1 args)
  when (not . null $ opts.errors) $
    throw (OptionsError opts.errors)
  (year, day_n) <- argumentsToPair opts.nonOptions
  -- If the verbose flag is set, hook up the logging writer to stderr
  let verbose = any (== Verbose) opts.options
  let logHandler : Eff [IO, Except Error, WriterL "log" String] () -> Eff [IO, Except Error] () =
    if verbose
       then handleLog ePutStrLn
       else handleLog (\x => pure ())
  -- Add the logging writer to the effect
  logHandler $ do
  -- Locate and read in the input file
  Just cwd <- currentDir | _ => throw NoCurrentDir
  let cwd = parse cwd
  let use_example = any (== UseExample) opts.options
  let base_dir = cwd /> "inputs" /> (if use_example then "examples" else "real");
  let input_path = show $ base_dir /> (show year) /> (show day_n)
  info "Reading input from \{input_path}"
  Right contents <- readFile input_path | Left err => throw $ InputRead input_path err
  -- Now add the reader that provides the input to the effect
  runReaderAt "input" contents {fs = PartEff Error} $ do
  -- Attempt to locate the provided day
  Just day <- pure $ locate year day_n advent | _ => throw $ NoSuchDay year day_n
  -- Run part 1
  part_1 <- lift $ runExcept day.part1
  (part_1, ctx) <- rethrow . mapLeft (SolveError year day_n 1 @{day.showErr1}) $ part_1
  putStrLn "\{show year} day \{show day_n} part 1:"
  putStrLn $ unlines . map ("  " ++) . lines . show @{day.showOut1} $ part_1
  -- Run part 2 if we have it
  case day.part2 of
    Nothing => pure ()
    Just part_2_f => do
      part_2 <- lift . runExcept $ part_2_f ctx
      part_2 <- rethrow . mapLeft (SolveError year day_n 2 @{day.showErr2}) $ part_2
      putStrLn "\{show year} day \{show day_n} part 2:"
      putStrLn $ unlines . map ("  " ++) . lines . show @{day.showOut2} $ part_2
  where
    -- print to standard error
    ePutStrLn : String -> IO ()
    ePutStrLn str = do
      _ <- fPutStrLn stderr str
      pure ()
    -- Decompose a writer, eases type inference
    unwriter : WriterL "log" String a -> (a, String)
    unwriter (Tell vw) = ((), vw)
    -- Lowers logging into IO within the effect using the given IO function
    handleLog :
      Has (WriterL "log" String) fs => Has IO (fs - WriterL "log" String) => 
      (tell : String -> IO ()) -> Eff fs a -> Eff (fs - (WriterL "log" String)) a
    handleLog tell x =
      handle
        (\msg, f =>
            do let (val, msg) = unwriter msg
               _ <- send $ tell msg
               f val) x

covering
||| Shim main, which simply executes the effectful computation
main : IO ()
main = runEff start [id, handleError]
  where
  handleError : Except Error a -> IO a
  handleError (Err err) = do
    printLn err
    exitFailure
